/* GLSL_multilight: different methods to display multiple lights, using plain GL, GLSL shaders or deferred shading
   Copyright (C) 2008 Angelo "Encelo" Theodorou
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glext.h> 

#include "init.h"
#include "extra.h"
#include "textfile.h"


void DrawScene(void);
void DrawLights(void);
void DrawQuad(void);

Light lights[3] = {
	{{ 0.0f, 2.0f,-2.0f, 1.0f}, { 0.0f, 0.0f, 0.0f, 1.0f},
		{ 1.0f, 0.0f, 0.0f, 1.0f}, { 1.0f, 1.0f, 1.0f, 1.0f}},
	{{-5.5f, 2.0f,-2.0f, 1.0f}, { 0.0f, 0.0f, 0.0f, 1.0f},
		{ 0.0f, 1.0f, 0.0f, 1.0f}, { 1.0f, 1.0f, 1.0f, 1.0f}},
	{{ 5.5f, 2.0f,-2.0f, 1.0f}, { 0.0f, 0.0f, 0.0f, 1.0f},
		{ 0.0f, 0.0f, 1.0f, 1.0f}, { 1.0f, 1.0f, 1.0f, 1.0f}}
};
GLuint texid[2];
struct Shaders sh;
int bg = 1;
GLfloat movz = 0.0f, roty = 0.0f;


int main(int argc, char **argv)
{
	SDL_Surface *screen;
	SDL_Event event;
	const SDL_VideoInfo* info = NULL;
	int width = 640;
	int height = 480;
	int bpp = 0;
	int flags = 0;

	const GLubyte* string;

	GLfloat black[] = {0.0f, 0.0f, 0.0f, 1.0f};
	GLfloat model_ambient[] = {0.2f, 0.2f, 0.2f, 1.0f};
	struct Material mat = {{0.6, 0.6, 0.6, 1.0}, {0.3, 0.3, 0.3, 1.0}, 100.0};

	struct FBO fbo;
	GLenum buffers[] = {GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_COLOR_ATTACHMENT2_EXT};
	GLint loc, multish = -1;

	int i, quit = 0;
	Uint32 interval;
	GLfloat movl = 0.0f;
	int rendermode = 0, movelights = 0, move = 0, rotate = 1;


/* ----- SDL init --------------- */
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		exit(-1);
	}

	atexit(SDL_Quit);
	
	info = SDL_GetVideoInfo();
	bpp = info->vfmt->BitsPerPixel;

	
/* ----- OpenGL attribute setting via SDL --------------- */
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	
/*	flags = SDL_OPENGL | SDL_FULLSCREEN; */
	flags = SDL_OPENGL;

	
/* ----- Setting up the screen surface --------------- */
	if ((screen = SDL_SetVideoMode(width, height, bpp, flags)) == 0) {
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
        	exit(-1);
	}

	SDL_WM_SetCaption("GLSL_multilight by Encelo", NULL);


/* ----- Checking for OpenGL 2 --------------- */
	string = glGetString(GL_VENDOR);
	printf("Vendor: %s\n", string);
	string = glGetString(GL_RENDERER);
	printf("Renderer: %s\n", string);
	string = glGetString(GL_VERSION);
	printf("OpenGL Version: %s\n", string);
	if (string[0] != '2') {
		printf("OpenGL 2 not supported!\n");
		exit(-1);
	}
	string = glGetString(GL_SHADING_LANGUAGE_VERSION);
	printf("GLSL Version: %s\n\n", string);


/* ----- OpenGL init --------------- */
	glShadeModel(GL_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL); /* Either set this or clear the Z-buffer at each scene rendering pass */
	glBlendFunc(GL_ONE, GL_ONE);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_RESCALE_NORMAL); /* Avoid improper lighting of the uniformly scaled brick room */

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 100.0f);
	glMatrixMode(GL_MODELVIEW);

	InitLighting(&mat);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, model_ambient);

	fbo.width = width;
	fbo.height = height;
	InitFBOs(&fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	InitShaders(&sh);
	glUseProgram(sh.p[5]);
	loc = glGetUniformLocation(sh.p[5], "Tex0");
	glUniform1i(loc, 0);
	loc = glGetUniformLocation(sh.p[5], "Tex1");
	glUniform1i(loc, 1);
	loc = glGetUniformLocation(sh.p[5], "Tex2");
	glUniform1i(loc, 2);
	glUseProgram(0);

	texid[0] = LoadTexture("tex0.jpg");
	texid[1] = LoadTexture("tex1.png");


/* ----- Event cycle --------------- */
	while (!quit) {
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_QUIT:
				quit = 1;
				break;
			case SDL_KEYDOWN:
				switch( event.key.keysym.sym ){
				case SDLK_ESCAPE:
				case SDLK_q:
					quit = 1;
					break;
				case SDLK_F1:
					SDL_WM_ToggleFullScreen(screen);
					break;
				case SDLK_m:
					movelights = !movelights;
					break;
				case SDLK_SPACE:
					move = !move;
					break;
				case SDLK_r:
					rotate = !rotate;
					break;
				case SDLK_b:
					bg = !bg;
					break;
				case SDLK_0:
					movz = 0.0f;
					roty = 0.0f;
					movl = 0.0f;
					break;
				case SDLK_1:
					glUseProgram(0);
					glDisable(GL_LIGHTING);
					rendermode = 0;
					printf("No lighting at all\n");
					break;
				case SDLK_2:
					glUseProgram(0);
					glEnable(GL_LIGHTING);
					rendermode = 0;
					printf("Per-vertex single-pass fixed pipeline lighting\n");
					break;
				case SDLK_3:
					glUseProgram(0);
					glEnable(GL_LIGHTING);
					multish = 0;
					rendermode = 1;
					printf("Per-vertex multi-pass fixed pipeline lighting\n");
					break;
				case SDLK_4:
					multish = sh.p[0];
					rendermode = 1;
					printf("Per-vertex multi-pass shader lighting\n");
					break;
				case SDLK_5:
					multish = sh.p[1];
					rendermode = 1;
					printf("Per-pixel multi-pass shader lighting\n");
					break;
				case SDLK_6:
					glUseProgram(sh.p[2]);
					rendermode = 0;
					printf("Per-vertex single-pass shader lighting\n");
					break;
				case SDLK_7:
					glUseProgram(sh.p[3]);
					rendermode = 0;
					printf("Per-pixel single-pass shader lighting\n");
					break;
				case SDLK_8:
					rendermode = 2;
					printf("Deferred shading lighting\n");
					break;
				}
			}
		}

		interval = FrameTiming();

		
/* ----- Blitting on the screen --------------- */
		if (move)
			movz += 0.0025 * interval;
		if (rotate)
			roty += 0.1 * interval;
		if (movelights)
			movl += 0.0025 * interval;

		lights[0].position[0] = sinf(movl) * 2.0f;
		lights[0].position[2] = -5 + cosf(movl) * 5.0f;
		lights[1].position[1] = 2 + sinf(movl) * 2.0f;
		lights[1].position[2] = -5 + cosf(movl) * 5.0f;
		lights[2].position[1] = 2 + sinf(movl) * 2.0f;
		lights[2].position[2] = -5 + cosf(movl) * 5.0f;

		SetLight(&lights[0], GL_LIGHT0);
		SetLight(&lights[1], GL_LIGHT1);
		SetLight(&lights[2], GL_LIGHT2);

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();
		glTranslatef(0.0f, -2.0f, -10.0f);
		glRotatef(15.0f, 1.0f, 0.0f, 0.0f);

		switch(rendermode) {
		case 0: /* Single-pass */
			DrawScene();
			break;
		case 1: /* Multi-pass */
			if (!multish) {
				for (i = 1; i < NUM_LIGHTS; i++)
					glDisable(GL_LIGHT0+i);
			}

			glUseProgram(multish);
			for (i = 0; i < NUM_LIGHTS; i++) {
				if (i > 0) {
					glLightModelfv(GL_LIGHT_MODEL_AMBIENT, black);
					glEnable(GL_BLEND);
				}
				SetLight(&lights[i], GL_LIGHT0); /* GL_LIGHT0 "recycling" */
				DrawScene();
			}

			glLightModelfv(GL_LIGHT_MODEL_AMBIENT, model_ambient);
			glDisable(GL_BLEND);

			if (!multish) {
				for (i = 1; i < NUM_LIGHTS; i++)
					glEnable(GL_LIGHT0+i);
			}
			break;
		case 2: /* Deferred lighting */
			/* First pass: scene rendering */
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo.frame);
			glDrawBuffers(3, buffers);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glUseProgram(sh.p[4]);
			DrawScene();

			/* Second pass: deferred lighting */
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
			glDrawBuffer(GL_BACK);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glActiveTexture(GL_TEXTURE0);
			glBindTexture(GL_TEXTURE_2D, fbo.texid[0]);
			glActiveTexture(GL_TEXTURE1);
			glBindTexture(GL_TEXTURE_2D, fbo.texid[1]);
			glActiveTexture(GL_TEXTURE2);
			glBindTexture(GL_TEXTURE_2D, fbo.texid[2]);
			glUseProgram(sh.p[5]);

			glDisable(GL_DEPTH_TEST); /* Prevent the z-buffer from being updated with quad depth values */
			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);

			for (i = 0; i < NUM_LIGHTS; i++) {
				if (i > 0) {
					glLightModelfv(GL_LIGHT_MODEL_AMBIENT, black);
					glEnable(GL_BLEND);
				}
				SetLight(&lights[i], GL_LIGHT0); /* Setting light position is influenced by the current MODELVIEW matrix */
				glPushMatrix();

				glLoadIdentity();
				DrawQuad();
				glPopMatrix();
			}
			glLightModelfv(GL_LIGHT_MODEL_AMBIENT, model_ambient);
			glDisable(GL_BLEND);

			glMatrixMode(GL_PROJECTION);
			glPopMatrix();
			glMatrixMode(GL_MODELVIEW);
			glEnable(GL_DEPTH_TEST);

			glActiveTexture(GL_TEXTURE0);

			break;
		}
		/* Drawing lights in traditional forward rendering, on top of everythign else, no matter current rendering mode */
		DrawLights();

		SDL_GL_SwapBuffers();		
		/* SDL_Delay(20);  */ /* Decomment this if you want 1/50th screen update */
	}

	
/* ----- Quitting --------------- */
	glDeleteTextures(2, texid);
	DeleteShaders(&sh);
	DeleteFBOs(&fbo);

	SDL_Quit();
	return 0;	
}


void DrawScene(void)
{
	int i;
	GLuint prog;

	/* Brick room */
	if (bg) {
		glBindTexture(GL_TEXTURE_2D, texid[0]);
		glPushMatrix();
		glTranslatef(0.0f, 6.0f, -10.0f);
		glScalef(10.0f, 10.0f, 10.0f);
		DrawCube(1);
		glPopMatrix();
	}

	/* Rotating Cubes */
	glBindTexture(GL_TEXTURE_2D, texid[1]);
	for (i = 0; i < 5; i++) {
		glPushMatrix();
		glTranslatef(0.0f, 0.0f, 3.0 * sinf(movz));
		glTranslatef(-2.0f, 0.0f, 0.0);
		glTranslatef(i * 2.0f, 0.0f, i * (-4.0));
		glRotatef(roty, 0.0f, 1.0f, 0.0f);
		DrawCube(0);
		glPopMatrix();
	}
}


void DrawLights(void)
{
	int i;
	GLuint prog;

	glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glGetIntegerv(GL_CURRENT_PROGRAM, &prog);
	glUseProgram(0);
	for (i = 0; i < NUM_LIGHTS; i++)
		DrawLight(&lights[i], 0.1f);
	glUseProgram(prog);
	glPopAttrib();
}


void DrawQuad(void)
{
	glBegin(GL_QUADS);
		glMultiTexCoord2f(GL_TEXTURE0, 0.0, 0.0);
		glMultiTexCoord2f(GL_TEXTURE1, 0.0, 0.0);
		glMultiTexCoord2f(GL_TEXTURE2, 0.0, 0.0);
		glVertex2f(-1.0f, -1.0f);
		glMultiTexCoord2f(GL_TEXTURE0, 1.0, 0.0);
		glMultiTexCoord2f(GL_TEXTURE1, 1.0, 0.0);
		glMultiTexCoord2f(GL_TEXTURE2, 1.0, 0.0);
		glVertex2f(1.0f, -1.0f);
		glMultiTexCoord2f(GL_TEXTURE0, 1.0, 1.0);
		glMultiTexCoord2f(GL_TEXTURE1, 1.0, 1.0);
		glMultiTexCoord2f(GL_TEXTURE2, 1.0, 1.0);
		glVertex2f(1.0f, 1.0f);
		glMultiTexCoord2f(GL_TEXTURE0, 0.0, 1.0);
		glMultiTexCoord2f(GL_TEXTURE1, 0.0, 1.0);
		glMultiTexCoord2f(GL_TEXTURE2, 0.0, 1.0);
		glVertex2f(-1.0f, 1.0f);
	glEnd();
}
